
// This software is written by HKJ from lygte-info.dk

//#define WITH_ETH
#define WITH_SER
#define SHARED_BUFFER


#define BUFFER_SIZE 100

#include "CmdProcessor.h"

#ifdef WITH_ETH
#include <SPI.h>
#include <EEPROM.h>
#include <Ethernet2.h>
#endif



#define EEPROM_IPMASK 60
#define EEPROM_IPGATE 64
#define EEPROM_IPADDRESS 68
#define EEPROM_IPMAC 72 // Only last two blocks of mac is changeable

#define MAC1 0xDE
#define MAC2 0xAD 
#define MAC3 0xBE
#define MAC4 0xEF

#define ethernetResetPin 8			// Use 8, pin 9 is a pwm pin
#define port 4010

#ifdef WITH_ETH
char cmdEthBuffer[BUFFER_SIZE];
byte cmdEthBufferIndex=0;
EthernetServer server(port);
EthernetClient client;
#endif

#ifdef WITH_SER

#ifdef WITH_ETH
#ifdef SHARED_BUFFER
#define cmdSerBuffer cmdEthBuffer
#define cmdSerBufferIndex cmdEthBufferIndex
#else
char cmdSerBuffer[BUFFER_SIZE];
byte cmdSerBufferIndex=0;
#endif
#else
char cmdSerBuffer[BUFFER_SIZE];
byte cmdSerBufferIndex=0;
#endif
#endif


void (*_handleCmds)(Stream *stream,char* cmd);
const char *_name;


const char *CRLF="\r\n";

CmdProcessor::CmdProcessor() {
}

CmdProcessor::CmdProcessor(const char* name,void (*handleCmds)(Stream *stream,char* cmd)) {
  init(name,handleCmds);
}

void CmdProcessor::init(const char* name,void (*handleCmds)(Stream *stream,char* cmd)) {
  _handleCmds = handleCmds;
  _name=name;  
#ifdef WITH_SER  
  Serial.begin(9600);  // For debugging and handling commands
#endif

#ifdef WITH_ETH    
  pinMode(ethernetResetPin, OUTPUT);
  digitalWrite(ethernetResetPin, LOW);
  delay(100);
  digitalWrite(ethernetResetPin, HIGH);
  delay(1000);


  byte ipName[] = { 8,8,8,8 };    
  byte ip[] = { EEPROM.read(EEPROM_IPADDRESS),EEPROM.read(EEPROM_IPADDRESS+1),EEPROM.read(EEPROM_IPADDRESS+2),EEPROM.read(EEPROM_IPADDRESS+3) };    
  byte ipGate[] = { EEPROM.read(EEPROM_IPGATE),EEPROM.read(EEPROM_IPGATE+1),EEPROM.read(EEPROM_IPGATE+2),EEPROM.read(EEPROM_IPGATE+3) };    
  byte subnet[] = { EEPROM.read(EEPROM_IPMASK),EEPROM.read(EEPROM_IPMASK+1),EEPROM.read(EEPROM_IPMASK+2),EEPROM.read(EEPROM_IPMASK+3) };    
  byte mac[] = { MAC1, MAC2, MAC3, MAC3, EEPROM.read(EEPROM_IPMAC), EEPROM.read(EEPROM_IPMAC+1)};
  if (ip[0]==255 || subnet[3]!=255) {
    byte ip[] = { 10,0,0,91 };    
    byte ipGate[] = { 10,0,0,1 };    
    byte subnet[] = { 255,255,255,0 };    
  }
   
  Ethernet.begin(mac, ip,ipName,ipGate,subnet);
  server.begin();
  
#ifdef WITH_SER    
  if (Serial) {
    Serial.print("Server is at ");
    Serial.print(Ethernet.localIP());
    Serial.print(" port ");
    Serial.println(port);
  }
#endif  
#endif
}

void cmdIdnQ(Stream *stream) {
  stream->print("HKJ,");
  stream->print(_name);
  stream->print(CRLF);
}


void cmdEsrQ(Stream *stream) {
  stream->print("0\r\n");
}

#ifdef WITH_ETH

void parseAndSaveIP(byte addr) {
  char *p=strtok(NULL,".");
  EEPROM.write(addr++,atoi(p));
  p=strtok(NULL,".");
  EEPROM.write(addr++,atoi(p));
  p=strtok(NULL,".");
  EEPROM.write(addr++,atoi(p));
  p=strtok(NULL,".");
  EEPROM.write(addr,atoi(p));  
}

void showIP(byte addr,Stream *stream) {
  stream->print(EEPROM.read(addr++));
  stream->print(".");
  stream->print(EEPROM.read(addr++));
  stream->print(".");
  stream->print(EEPROM.read(addr++));
  stream->print(".");
  stream->print(EEPROM.read(addr++));
  stream->print(CRLF);
}

void cmdIPMask() {
  parseAndSaveIP(EEPROM_IPMASK);
}

void cmdIPMaskQ(Stream *stream) {
  showIP(EEPROM_IPMASK,stream);
}

void cmdIPAddress() {
  parseAndSaveIP(EEPROM_IPADDRESS);
}

void cmdIPAddressQ(Stream *stream) {
  showIP(EEPROM_IPADDRESS,stream);  
}

void cmdIPGate() {
  parseAndSaveIP(EEPROM_IPGATE);
}

void cmdIPGateQ(Stream *stream) {
  showIP(EEPROM_IPGATE,stream);  
}

void cmdIPMac() {
  char *p=strtok(NULL,".:-");
  byte addr=EEPROM_IPMAC;
  EEPROM.write(addr++,atoi(p));
  p=strtok(NULL,".:-");
  EEPROM.write(addr++,atoi(p));
}

void cmdIPMacQ(Stream *stream) {
  stream->print(MAC1);
  stream->print('-');
  stream->print(MAC2);
  stream->print('-');
  stream->print(MAC3);
  stream->print('-');
  stream->print(MAC4);
  stream->print('-');
  stream->print(EEPROM.read(EEPROM_IPMAC));
  stream->print('-');
  stream->print(EEPROM.read(EEPROM_IPMAC+1));
  stream->print(CRLF);
}

#endif

PROGMEM const char cmds[] ="*IDN?\0*ESR?\0FREEMEM?\0"
#ifdef WITH_ETH
  "IPMASK\0IPMASK?\0IPADDRESS\0IPADDRESS?\0IPGATE\0IPGATE?\0IPMAC\0IPMAC?\0"
#endif  
  "\0\0";


void freeRam (Stream *stream) {
  extern int __heap_start, *__brkval; 
  int v; 
  stream->print((int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval)); 
  stream->print(CRLF);
}


void processCmd(char *buffer,Stream *stream) {
  char *p=buffer;
  while (*p!=0 && *p!=' ') {		// Only first word is forced to uppercase
    *p=toupper(*p);
     p++;
  }

  p=buffer;
  while (p!=NULL && *p!=0) {  
    char *cmd1Buffer=p;
    p=strchr(p,';');
    if (p!=NULL) {
      *p=0;
      p++;
    }
    char *cmd=strtok(cmd1Buffer," ");
    if (*cmd==':') cmd++;
    switch (cmdidx(cmd,cmds)) {    
      case 0 : _handleCmds(stream,cmd);break;
      case 1 : cmdIdnQ(stream);break;
      case 2 : cmdEsrQ(stream);break;
      case 3 : freeRam(stream);break;
#ifdef WITH_ETH      
      case 4 : cmdIPMask();break;
      case 5 : cmdIPMaskQ(stream);break;
      case 6 : cmdIPAddress();break;
      case 7 : cmdIPAddressQ(stream);break;
      case 8 : cmdIPGate();break;
      case 9 : cmdIPGateQ(stream);break;
      case 10 : cmdIPMac();break;
      case 11: cmdIPMacQ(stream);break;
#endif      
    }    
   }
}

#ifdef WITH_ETH
void processEthRequest() {
  if (!client) return;
  if (!client.connected()) return;
  while (client.available()>0) {
    char c=client.read();
    if (c=='\r' || c=='\n') {
      cmdEthBuffer[cmdEthBufferIndex]=0;
      processCmd(cmdEthBuffer,&client);
      cmdEthBufferIndex=0;
    } else if (c!=0) {
      cmdEthBuffer[cmdEthBufferIndex]=c;
      if (cmdEthBufferIndex<BUFFER_SIZE) cmdEthBufferIndex++;
    }
  }
}
#endif

#ifdef WITH_SER  
void processSerRequest() {
  // if (!Serial) return;			// This takes 10ms
  while (Serial.available()>0) {		
    char c=Serial.read();
    if (c=='\r' || c=='\n') {
      cmdSerBuffer[cmdSerBufferIndex]=0;
      processCmd(cmdSerBuffer,&Serial);
      cmdSerBufferIndex=0;
    } else if (c!=0) {
      cmdSerBuffer[cmdSerBufferIndex]=c;
      if (cmdSerBufferIndex<BUFFER_SIZE) cmdSerBufferIndex++;
    }
  }
}
#endif
void CmdProcessor::processInput() {  
#ifdef WITH_ETH
  client = server.available();
  if (client) {
    processEthRequest();
  }
#endif  
#ifdef WITH_SER  
  processSerRequest();
#endif  
}  

void CmdProcessor::processCmd(char *buffer,Stream *stream) {  
  ::processCmd(buffer,stream);
}  

  

byte cmdidx(const char *cmd,PGM_P cmd_list) {
  byte idx=1;
  while (pgm_read_byte_near(cmd_list)!=0) {
    const char *p=cmd;
    byte fail=0;
    char c;
    while ((c=pgm_read_byte_near(cmd_list))!=0) {
       if (c!=toupper(*p)) fail=1;
       cmd_list++;
       p++;
       }
      if (!fail && *p==0) return idx; 
      idx++; 
      cmd_list++;
  }
  return 0;
}
